import torch
from torch_sparse import SparseTensor
from torch_geometric.nn import SignedConv


def test_signed_conv():
    x = torch.randn(4, 16)
    edge_index = torch.tensor([[0, 1, 2, 3], [0, 0, 1, 1]])
    row, col = edge_index
    adj = SparseTensor(row=row, col=col, sparse_sizes=(4, 4))

    conv1 = SignedConv(16, 32, first_aggr=True)
    assert conv1.__repr__() == 'SignedConv(16, 32, first_aggr=True)'

    conv2 = SignedConv(32, 48, first_aggr=False)
    assert conv2.__repr__() == 'SignedConv(32, 48, first_aggr=False)'

    out1 = conv1(x, edge_index, edge_index)
    assert out1.size() == (4, 64)
    assert conv1(x, adj.t(), adj.t()).tolist() == out1.tolist()

    out2 = conv2(out1, edge_index, edge_index)
    assert out2.size() == (4, 96)
    assert conv2(out1, adj.t(), adj.t()).tolist() == out2.tolist()

    t = '(Tensor, Tensor, Tensor) -> Tensor'
    jit1 = torch.jit.script(conv1.jittable(t))
    jit2 = torch.jit.script(conv2.jittable(t))
    assert jit1(x, edge_index, edge_index).tolist() == out1.tolist()
    assert jit2(out1, edge_index, edge_index).tolist() == out2.tolist()

    t = '(Tensor, SparseTensor, SparseTensor) -> Tensor'
    jit1 = torch.jit.script(conv1.jittable(t))
    jit2 = torch.jit.script(conv2.jittable(t))
    assert jit1(x, adj.t(), adj.t()).tolist() == out1.tolist()
    assert jit2(out1, adj.t(), adj.t()).tolist() == out2.tolist()

    adj = adj.sparse_resize((4, 2))
    assert torch.allclose(conv1((x, x[:2]), edge_index, edge_index), out1[:2],
                          atol=1e-6)
    assert torch.allclose(conv1((x, x[:2]), adj.t(), adj.t()), out1[:2],
                          atol=1e-6)
    assert torch.allclose(conv2((out1, out1[:2]), edge_index, edge_index),
                          out2[:2], atol=1e-6)
    assert torch.allclose(conv2((out1, out1[:2]), adj.t(), adj.t()), out2[:2],
                          atol=1e-6)

    t = '(PairTensor, Tensor, Tensor) -> Tensor'
    jit1 = torch.jit.script(conv1.jittable(t))
    jit2 = torch.jit.script(conv2.jittable(t))
    assert torch.allclose(jit1((x, x[:2]), edge_index, edge_index), out1[:2],
                          atol=1e-6)
    assert torch.allclose(jit2((out1, out1[:2]), edge_index, edge_index),
                          out2[:2], atol=1e-6)

    t = '(PairTensor, SparseTensor, SparseTensor) -> Tensor'
    jit1 = torch.jit.script(conv1.jittable(t))
    jit2 = torch.jit.script(conv2.jittable(t))
    assert torch.allclose(jit1((x, x[:2]), adj.t(), adj.t()), out1[:2],
                          atol=1e-6)
    assert torch.allclose(jit2((out1, out1[:2]), adj.t(), adj.t()), out2[:2],
                          atol=1e-6)
